"""
@Author: Simona Bernardi, Raúl Javierre
@Date: 07/05/2022

teg module Version 2.0.0
This modules includes the following classes:

- TEGDetector (API class)
- LevelExtractor
- TEGGenerator
- GraphDistanceCollector
- ModelBuilder
- AnomalyDetector

that implements the detectors based on Time Evolving Graph (TEG) and graph dissimilarity distribution.

"""

from time import time
import numpy as np
import pandas as pd
import os
import sys      

from tegdet.graph_comparison import *


class TEGDetector():
    """ 
    API 
    """
    #Default values
    __N_BINS = 30
    __N_OBS_PER_PERIOD = 336
    __ALPHA = 5

    def __init__(self, metric, n_bins=__N_BINS, n_obs_per_period=__N_OBS_PER_PERIOD, alpha=__ALPHA):
        self.__metric = metric
        self.__n_bins = n_bins
        self.__n_obs_per_period= n_obs_per_period
        self.__alpha = alpha

    def get_dataset(self,ds_path):
        """
        Load the dataset from "ds_path" (csv file), rename the two columns as TS (timestamp) and 
        data points (DP), respectively, and return the pandas dataframe "dt"
        """
        df = pd.read_csv(ds_path)
        df.columns = ['TS','DP']
        return df

    def build_model(self, training_dataset):
        """
        Build the prediction model based on the "training_dataset" and return it together with
        the time to build the model
        """
        t0 = time()
        obs = training_dataset['DP']
        mb = ModelBuilder(obs, self.__n_bins)
        mb.build_model(self.__metric, int(len(training_dataset.index) / self.__n_obs_per_period))

        return mb, time() - t0

    def predict(self, testing_dataset, model):
        """
        Make predictions on the "testing_dataset" using the model. It returns three values:
        number of outliers, total number of observations, and time to make predictions
        """
        t0 = time()
        data_points = testing_dataset['DP']
        ad = AnomalyDetector(model)
        test = ad.make_prediction(self.__metric, data_points, int(len(testing_dataset.index) / self.__n_obs_per_period))
        outliers = ad.compute_outliers(test, 100 - self.__alpha)

        return outliers, int(len(testing_dataset.index) / self.__n_obs_per_period), time() - t0
        
    def compute_confusion_matrix(self, ground_true, predictions):
        """
        Pre: "ground_true" is a vector with the true values, "predictions" is a vector 
        with predicted values (0,1)
        Post: Computes the confusion matrix. It returns the confusion matrix as dictionary type.
        """
        cm = {'tp': 0, 'tn': 0, 'fp': 0, 'fn': 0}
        cm['tp'] = ((ground_true == 1) & (predictions == 1)).sum()
        cm['tn'] = ((ground_true == 0) & (predictions == 0)).sum()
        cm['fp'] = ((ground_true == 0) & (predictions == 1)).sum()
        cm['fn'] = ((ground_true == 1) & (predictions == 0)).sum()

        return cm

    def print_metrics(self, detector, testing_set, perf, cm):
        """
        Print on the stdout: the "detector" configuration, the "testing set", the performance metrics "perf" 
        (time to build the model and to make predictions) and the confusion matrix "cm" 
        """
        print("Detector:\t\t\t", detector['metric'])
        print("N_bins:\t\t\t\t", detector['n_bins'])
        print("N_obs_per_period:\t\t", detector['n_obs_per_period'])
        print("Alpha:\t\t\t\t", detector['alpha'])
        print("Testing set:\t\t\t", testing_set)
        print("Time to build the model:\t", perf['tmc'], "seconds")
        print("Time to make prediction:\t", perf['tmp'], "seconds")
        print("Confusion matrix:\t\n\n", cm)

    def metrics_to_csv(self, detector, testing_set, perf, cm,results_csv_path):
        """
        Save in the csv file "result_csv_path", the "detector" configuration, the  "testing set",
        the performance metrics "perf" (time to build the model and to make predictions)
        and the confusion matrix "cm" 
        """
        df = pd.DataFrame({
                           'detector': detector['metric'],
                           'n_bins': detector['n_bins'],
                           'n_obs_per_period': detector['n_obs_per_period'],
                           'alpha': detector['alpha'],
                           'testing_set': testing_set,
                           'time2build': perf['tmc'],
                           'time2predict': perf['tmp'],
                           'tp': cm['tp'],
                           'tn': cm['tn'],
                           'fp': cm['fp'],
                           'fn': cm['fn']},
                           index=[0])

        df.to_csv(results_csv_path, mode='a', header=not os.path.exists(results_csv_path), index=False)


class LevelExtractor:
    """ 
    Extractor of levels, univariate time-series discretizer 
    """

    def __init__(self, min_value, step, n_bins):
        """
        Create levels [0,1,..,n_bins+2], the last two positions for the possible 
        outliers of the testing dataset (minimum than the min_train_value and maximum 
        of the max_train_value)
        """
        self.__levels = np.arange(n_bins+2)
        self.__step = step
        self.__min_value = min_value

    def get_levels(self):
        return self.__levels

    def discretize(self, observations):
        """
        Discretization of  "observations" according to the "self.levels"
        "observations" is a np.array (of floats)
        """
        n_obs = len(observations)  # number of observations
        discretized_obs = -1 * np.ones(n_obs)  # array initialized with -1
        discretized_obs = discretized_obs.astype(int)  # np.array of int

        # Case: "observations" (testing set) is lower than the min_train_value
        #       level position the last
        discretized_obs = np.where((observations < self.__min_value), self.__levels[-1], discretized_obs)
        
        n_bins= len(self.__levels)-2
        i = 0  # while iterator
        while i < n_bins:
            lower = self.__min_value + i * self.__step
            upper = self.__min_value + (i + 1) * self.__step
            discretized_obs = np.where((lower <= observations) & (observations < upper),
                             self.__levels[i], discretized_obs)
            i += 1

        # Case: "observations" (testing set) is greater than the max_train_value 
        # level position the penultimate      
        discretized_obs = np.where(upper <= observations, self.__levels[-2], discretized_obs)

        return discretized_obs


class TEGGenerator:
    """
    Time Evolving Graph generator
    """

    def __init__(self, observations_discretized, n_periods):
        """
        Generates and set the time evolving graphs series from the discretized observations
        "observation_discretized" and the number of periods "n_periods"
        """
        n_obs = int(len(observations_discretized) / n_periods)  # number of observations per period
        self.__teg = []
        for period in range(n_periods):
            obs_discr_period = observations_discretized[period * n_obs:(period + 1) * n_obs]
            # Transforms to a dataframe (needed to generate the graph)
            df = pd.DataFrame({'Period': period * np.ones(n_obs), 'DP': obs_discr_period})
            gr = Graph()
            gr.generate_graph(df)
            self.__teg.append(gr)

    def get_teg(self):

        return self.__teg


class GraphDistanceCollector:
    """
    Collector of distances between graphs in teg and the global graph
    """

    def __init__(self, n_periods):

        
        self.__distance = np.empty(n_periods)
 
    def compute_graphs_dist(self, teg, global_graph, metric):
        """
        Compute and return the distances between  each graph in "teg" and the "global_graph"
        using the dissimilarity "metric"
        """
        
        gc_name = "Graph" + metric + "Dissimilarity"

        for period in range(self.__distance.size):
            #Create an instance of the metric specific graph comparator
            #by converting a string to a class object
            gc = getattr(sys.modules[__name__], gc_name)(teg[period], global_graph)
            #Graph resizing
            gc.resize_graphs()
            #Compute the difference based on the metric
            self.__distance[period] = gc.compare_graphs()
        
        return self.__distance

class ModelBuilder:
    """
    Prediction model builder based on teg and baseline graph dissimilarity distribution
    """

    def __init__(self, observations, n_bins):
        """
        Constructor that initializes attributes based on the training dataset "observations" and "n_bins"
        Creation of a new level extractor.
        """
        m = observations.min()  # min of the TRAINING dataset
        M = observations.max()  # max of the TRAINING dataset
        step = (M - m) / n_bins  # usage increment step
        self.__obs = observations
        self.__le = LevelExtractor(m, step, n_bins)
        self.__baseline = None
        self.__global_graph= None

    def get_level_extractor(self):
        return self.__le

    def get_baseline(self):
        return self.__baseline

    def get_global_graph(self):
        return self.__global_graph

    def __sum_graphs(self, gr1, gr2):
        """
        Pre: Graph "gr1" nodes set includes graph "gr2" nodes set
        Post: Added the graph "gr2" to graph "gr1" 
        """
        nodes = gr2.get_nodes()
        nodes_freq = gr2.get_nodes_freq()
        matrix = gr2.get_matrix()        
        n_nodes = nodes.size
        for i in range(n_nodes):
            row = nodes[i]
            gr1.update_node_freq(row, nodes_freq[i])
            for j in range(n_nodes):
                col = nodes[j]
                gr1.update_matrix_entry(row,col,matrix[i][j])

    def __compute_global_graph(self, graphs):
        """
        Create and return a global graph as the sum of a list of "graphs".
        """
        n_bins = len(self.__le.get_levels())  
        self.__global_graph = Graph(np.arange(n_bins, dtype=int), np.zeros((n_bins), dtype=int), np.zeros((n_bins, n_bins), dtype=int))      
        
        for gr in graphs:
            self.__sum_graphs(self.__global_graph, gr)


    def build_model(self, metric, n_periods):
        """
        Build the distribution of the dissimilarities based on "metric", "observations", and
        "n_periods". It returns the ditribution of the dissimilarities and the global graph.
        """

        # Get observation levels (discretization)
        obs_discretized = self.__le.discretize(self.__obs)

        # Get the time-evolving graphs
        tegg = TEGGenerator(obs_discretized, n_periods)
        graphs = tegg.get_teg()

        # Get the global graph of the training period
        self.__compute_global_graph(graphs)

        # Get the graph dissimilarity baseline distribution
        gdc = GraphDistanceCollector(n_periods)
        self.__baseline = gdc.compute_graphs_dist(graphs, self.__global_graph, metric)


class AnomalyDetector:
    """
    Make predictions and compute outliers 
    """

    def __init__(self, model):

        self.__model = model

    def make_prediction(self, metric, observations, n_periods):
        """
        Make and return the predictions of the "observations" based on the "baseline" distribution 
        of the dissimilarities, "global_graph", the dissimilarity "metric" and "n_periods"
        """

        # Gets consumption levels (consumption discretization)
        obs_discretized = self.__model.get_level_extractor().discretize(observations)

        # Generates the time-evolving graphs
        tegg = TEGGenerator(obs_discretized, n_periods)
        graphs = tegg.get_teg()
 
        # Computes the distance between each graph and the global graph        
        global_graph = self.__model.get_global_graph()
        gdc = GraphDistanceCollector(n_periods)
        graph_dist = gdc.compute_graphs_dist(graphs, global_graph, metric)
        
        return graph_dist

    def compute_outliers(self, prediction, sigLevel):
        """
        Compute the outliers based on the baseline graph dissimilarity distribution,
        the "prediction" and the significance level "sigLevel"
        """
        baseline = self.__model.get_baseline()
        perc = np.percentile(baseline, sigLevel)

        # Sets a counter vector to zero
        n_out = np.zeros(prediction.size)
        # Dissimilarity tests
        n_out = np.where(prediction > perc, n_out + 1, n_out)

        return n_out







"""
@Author: Simona Bernardi
@Date: updated 02/05/2022

Input dataset:
- energy consumption (in KhW), every half-an-hour, registered by a smartmeter.
- the training set is over 60 weeks
- the testing sets are over 15 weeks: a testing set is either normal or anomalous testings
- The training set and testing sets are stored in separate files in the "/dataset/" folder.

The test program:
- builds a prediction model based on the training set for each dissimilarity metric (TEG-detectors variants)
- makes predictions in testing sets and
- compute the confusion matrix and performance metrics  

Output:
- Confusion matrices and performance metrics are stored in RESULTS_PATH
- Validation: results are compared with the ones obtained from the TEG version of the Diaspore repository
(assumed correct), which are stored in REFERENCE_PATH
"""

import os
import pandas as pd
import numpy as np

from tegdet.teg import TEGDetector


#Input datasets/output results Paths
TRAINING_DS_PATH = "/dataset/training.csv"
TEST_DS_PATH = "/dataset/test_"
RESULTS_PATH = "/script_results/test_detector_comparer_TEG_results.csv"
REFERENCE_PATH = "/script_results/reference_results.csv"

#List of testing
list_of_testing = ("normal", "anomalous")

#List of metrics (detector variants)
list_of_metrics = [ "Cosine",  "Jaccard", "Hamming", "KL", "Jeffreys", "JS", "Euclidean", 
                    "Cityblock", "Chebyshev", "Minkowski", "Braycurtis", "Kulczynski", 
                    "Canberra", "Bhattacharyya", "Squared", "Divergence", "Additivesymmetric"]

#Parameters of TEG detectors: default values
n_bins = 30
n_obs_per_period=336
alpha=5


def test_generate_results():

    cwd = os.getcwd() 

    train_ds_path = cwd + TRAINING_DS_PATH

    for metric in list_of_metrics:

        teg = TEGDetector(metric)

        #Load training dataset
        train_ds = teg.get_dataset(train_ds_path)

        assert not train_ds.empty, "The training dataset is empty."

        
        #Build model
        model, time2build = teg.build_model(train_ds)

        for testing in list_of_testing:

            #Path of the testing
            test_ds_path = cwd + TEST_DS_PATH + testing + ".csv"
                
            #Load testing dataset
            test_ds = teg.get_dataset(test_ds_path)

            assert not test_ds.empty, "The testing dataset is empty."

            #Make prediction
            outliers, obs, time2predict = teg.predict(test_ds, model)

            #Set ground true values
            if testing == "anomalous":
                groundtrue = np.ones(obs)        
            else:
                groundtrue = np.zeros(obs)

            #Compute confusion matrix
            cm = teg.compute_confusion_matrix(groundtrue, outliers)

            #Collect detector configuration
            detector = {'metric': metric, 'n_bins': n_bins, 'n_obs_per_period':n_obs_per_period, 'alpha': alpha}

            #Collect performance metrics in a dictionary
            perf = {'tmc': time2build, 'tmp': time2predict}

            #Print and store basic metrics
            teg.print_metrics(detector, testing, perf, cm)
            results_path = cwd + RESULTS_PATH
            teg.metrics_to_csv(detector, testing, perf, cm, results_path)

        assert os.path.exists(results_path), "Results file has not been created."
        assert os.path.getsize(results_path) > 0, "The result file is empty"


def test_results():

    #Get current working directory
    cwd = os.getcwd() 

    #Load results and reference results
    results_path = cwd + RESULTS_PATH
    reference_path = cwd + REFERENCE_PATH
    results = pd.read_csv(results_path)
    reference = pd.read_csv(reference_path)
    
    #Select confusion matrix entries
    cm = ['tp','tn','fp','fn']
    results = results[cm]
    reference = reference[cm]

    #Compare results with  the reference results
    assert results.equals(reference), "Results are not equal to the reference results."

if __name__ == '__main__':

    test_generate_results()

    test_results()





from setuptools import find_packages, setup

#Read the contents of the README file
with open("README.txt", "r") as fh:
    long_description = fh.read()

setup(
    name='tegdet',
    packages=find_packages(include=['tegdet']),
    version='1.0.0',
    description='Time Evolving Graph detectors library',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/DiasporeUnizar/TEG",
    author='Diaspore team',
    author_email="simonab@unizar.es",
    license='GPL-2',
    classifiers=[
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Operating System :: OS Independent"
    ],
    python_requires='>=3.6.1',
    setup_requires=['pytest-runner'],
    tests_require=['pytest>=6.2.2'],
    test_suite='test'
)


"""
@Author: Simona Bernardi
@Date: 31/05/2022

graph_comparison module Version 2.0.0

- Graph class enables to generate a causal graph (node frequency list, adjacency-frequency matrix)
from the dataset
- GraphComparator and sub-classes enable to compare two graphs and compute the "difference" between them according to a 
given measure

"""

import numpy as np

from abc import ABC, abstractmethod
from math import sqrt
from scipy.stats import entropy #it is used to compute the KLD measure
from scipy.spatial import distance #it is used to compute several distances

class Graph:
    """
    Graph generator and manipulator (graph expansion)
    """

    def __init__(self, nodes=None, nodes_freq=None, matrix=None):
        """
        Constructor that initializes the graph attributes
        """       
        self.__nodes = nodes          
        self.__nodes_freq = nodes_freq
        self.__matrix = matrix

    def get_nodes(self):
        return self.__nodes

    def get_nodes_freq(self):
        return self.__nodes_freq

    def get_matrix(self):
        return self.__matrix

    def update_node_freq(self, pos, value):
        self.__nodes_freq[pos] += value

    def update_matrix_entry(self, row, col, value):
        self.__matrix[row][col] += value

    def __get_index(self, element):
        """
        Returns the index of the matrix row (column) based on "element"
        """
        idx = -1  # not assigned
        i = 0
        while i < len(self.__nodes) and idx == -1:
            if element == self.__nodes[i]:
                idx = i
            i += 1

        return idx
   
    def generate_graph(self, obs_discretized):
        """
        Generates the graph from the discretized observations "obs_discretized"
        """
        grouped = obs_discretized.groupby('DP').count()
        # Sets vertices: they are ordered according to the levels 
        self.__nodes = grouped.index.to_numpy()          
        self.__nodes_freq = grouped.to_numpy()
        dim = len(self.__nodes)

        # Initializes the adjacent matrix
        self.__matrix = np.zeros((dim, dim), dtype=int)
        attr = obs_discretized.DP.to_numpy()
        # Sets the adjacent matrix with the frequencies
        for i in range(attr.size - 1):
            row = self.__get_index(attr[i])
            col = self.__get_index(attr[i + 1])
            self.__matrix[row][col] += 1

    def expand_graph(self, position, vertex):
        """
        Expands the graph by inserting a new node "vertex" in "position". The new added fictious node
        has frequency -1. The new added row and column of the adjacency matrix have -1 entries
        """
        # Different from zero to differentiate from the absence of arc, but presence of the node
        wildcard = '-1'
        # Insert the new vertex in the list of nodes
        self.__nodes = np.insert(self.__nodes, position, vertex)
        self.__nodes_freq = np.insert(self.__nodes_freq, position, wildcard)
        # Insert the new column in the matrix
        self.__matrix = np.insert(self.__matrix, position, wildcard, axis=1)
        # Insert the new row in the matrix
        self.__matrix = np.insert(self.__matrix, position, wildcard, axis=0)


class GraphComparator(ABC):
    """ 
    Graphs comparator operator (abstract class)
    """
    def __init__(self, first_graph, second_graph):
        """
        Constructor that initializes the two operands
        """
        self._graph1 = first_graph
        self._graph2 = second_graph

    def _normalize_matrices(self):
        """
        Flatten the matrices of the two graphs and normalize them
        """
        # Get the two matrices and convert them into arrays
        edges1 = self._graph1.get_matrix().flatten()
        edges2 = self._graph2.get_matrix().flatten()

        # Set -1 entries to zero
        edges1 = np.where((edges1 < 0), edges1 * 0, edges1)
        edges2 = np.where((edges2 < 0), edges2 * 0, edges2)

        # Normalizes the matrices (PDF)
        edges1 = edges1 / (edges1.sum())
        edges2 = edges2 / (edges2.sum())

        return edges1, edges2

    def resize_graphs(self):
        """
        Compare the nodes of the two graphs and possibly expand them
        """

        # Union of the nodes
        union = np.union1d(self._graph1.get_nodes(), self._graph2.get_nodes())

        # Compare the node list and possibly extend the graph(s)
        for i in range(union.size):
            nodes = self._graph1.get_nodes()
            if (nodes.size > i) and (nodes[i] != union[i]) or (nodes.size <= i):
                self._graph1.expand_graph(i, union[i])

        for i in range(union.size):
            nodes = self._graph2.get_nodes()
            if (nodes.size > i) and (nodes[i] != union[i]) or (nodes.size <= i):
                self._graph2.expand_graph(i, union[i])

    @abstractmethod
    def compare_graphs(self):  # signature only because it is overriden
        pass



#######################################################################
# Graph Edit Distance (GED) family --> this family mostly used in TEG
#######################################################################
class GraphHammingDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: Structural-based distance (Hamming)
        The two matrix arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Get just the matrices and convert into arrays
        first = self._graph1.get_matrix().flatten()
        second = self._graph2.get_matrix().flatten()
        # Just the adjacency matrix
        dim = min(len(first), len(second))
        # Setting a counter vector to zero
        counter = np.zeros(dim)
        # Count if both elements are either positive or zero
        counter = np.where(((first > 0) & (second > 0)) |
                           ((first == 0) & (second == 0)), counter + 1, counter)

        distance = 1.0 - np.sum(counter) / float(dim)

        return distance  #returns the dissimilarity (distance)

'''
Reference of the following implemented dissimilarity metrics:
- Cha, "Comprehensive Survey on Distance/Similarity Measures between Probability Density Functions", 2007
- See also URL: https://cran.r-project.org/web/packages/philentropy/vignettes/Introduction.html
'''

####################################################################
# Inner product family (3 different distance metrics)
####################################################################

class GraphCosineDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: Frequency-based dissimilarity (1-cosine)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Convert into arrays the node frequencies and matrices
        first = np.concatenate(
            (self._graph1.get_nodes_freq(), self._graph1.get_matrix().flatten()), axis=None)
        second = np.concatenate(
            (self._graph2.get_nodes_freq(), self._graph2.get_matrix().flatten()), axis=None)
    
        # Normalization factor
        nfactor = 1.0
        sp = first * second / nfactor
        # Frobenius norm (L2-norm Euclidean)
        norm1 = np.linalg.norm(first)
        norm2 = np.linalg.norm(second)
        den = np.sum(sp)
        
        if den > 0:
            # Compute the product
            cosinus =  den / (norm1 * norm2)
        else:
            cosinus = 0

        #Since some entries of the matrices can be -1 the cosinus maybe be negative!
        return 1.0 - cosinus #returns the dissimilarity (distance)

class GraphJaccardDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Jaccard)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """
        # Matrices normalization
        first, second = self._normalize_matrices()
 
        # Compute the Jaccard similarity (equal to Peak Correlation Energy)
        sumprod = (first * second).sum()
        quadnorm1 = (first*first).sum()
        quadnorm2 = (second*second).sum()
        jac = sumprod / (quadnorm1 + quadnorm1 - sumprod)

        return 1.0 - jac #returns the dissimilarity (distance)

class GraphDiceDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Dice)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """
        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Dice similarity 
        sumprod = (first * second).sum()
        quadnorm1 = (first*first).sum()
        quadnorm2 = (second*second).sum()
        dice = (2 * sumprod) / (quadnorm1 + quadnorm1)

        return 1.0 - dice #returns the dissimilarity (distance)


####################################################################
# Shannon's entropy family 
####################################################################
class GraphKLDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Kullback–Leibler)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """
        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the KLD of first  w.r.t second 
        kld = entropy(first,second,base=2)

        return kld  #returns the dissimilarity (distance)

class GraphJeffreysDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Jeffreys)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """
        # Matrices normalization
        first, second = self._normalize_matrices()


        # Compute the Jeffreys of first  w.r.t second 
        
        # Check possible case of division by 0 
        almostzero = np.finfo(float).eps
        second = np.where( (second < almostzero), almostzero, second)

        # Check possible case of log 0 
        jef = np.where((first < almostzero), almostzero, first / second)
        jef = np.log(jef) * (first - second)

        return jef.sum() #returns the dissimilarity (distance)

class GraphJSDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic:  PDF-based dissimilarity (Jensen-Shannon)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """
        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the JSD 
        jsd = distance.jensenshannon(first,second,base=2)

        return jsd  #returns the dissimilarity (distance)

###################################################################
# Minkowski family (4 different distance metrics)
###################################################################
class GraphEuclideanDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Euclidean)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Euclidean distance 
        eucl = distance.euclidean(first,second)

        return eucl #returns the dissimilarity (distance)

class GraphCityblockDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Cityblok)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Cityblock distance 
        city = distance.cityblock(first,second)

        return city  #returns the dissimilarity (distance)

class GraphChebyshevDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Chebyshev) 
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Chebyshev distance 
        cheb = distance.chebyshev(first,second)

        return cheb  #returns the dissimilarity (distance)

class GraphMinkowskiDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Minkowski distance with p=3)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Minkowski distance 
        mink = distance.minkowski(first,second,3)

        return mink  #returns the dissimilarity (distance)

###################################################################
# L_1 family (6 different distance metrics) 
###################################################################
class GraphBraycurtisDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Bray-Curtis, also called Sorensen)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
        graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Bray-Curtis distance 
        bray = distance.braycurtis(first,second)

        return bray  #returns the dissimilarity (distance)

class GraphGowerDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Gower)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Gower distance (=Cityblock divided by the number of elements)
        gower = distance.cityblock(first,second) / len(first)

        return gower  #returns the dissimilarity (distance)

class GraphSoergelDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Soergel)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Soergel distance (=Cityblock divided by the 
        # sum of the pairwise_max_elements)
        soergel = distance.cityblock(first,second) / np.maximum(first,second).sum()

        return soergel  #returns the dissimilarity (distance)

class GraphKulczynskiDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Kulczynski)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Kulczynski distance (=Cityblock divided by the 
        # sum of the pairwise_min_elements)
        num = distance.cityblock(first,second)
        den = np.minimum(first,second).sum()
        
        #Check possible division by zero 
        almostzero = np.finfo(float).eps
        if den > almostzero:
            kulc =  num / den
        else:    
            kulc = num / almostzero

        return kulc  #returns the dissimilarity (distance)

class GraphCanberraDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Canberra)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Canberra distance 
        canb = distance.canberra(first,second)

        return canb  #returns the dissimilarity (distance)

class GraphLorentzianDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Lorentzian)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Lorentzian distance 
        lore = np.log(1 + (first - second))

        return lore.sum()  #returns the dissimilarity (distance)

###################################################################
# Fidelity or Squared-chord family (4 different distance metrics) 
###################################################################
class GraphBhattacharyyaDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Bhattacharyya)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Bhattacharyya distance 
        bhatta = (np.sqrt(first * second)).sum()
        
        #Check possible case of log(0) 
        almostzero = np.finfo(float).eps
        if bhatta > almostzero:
            bhatta = -1.0 * np.log(bhatta)
        else:   
            bhatta = -1.0 * np.log(almostzero)

        return  bhatta #returns the dissimilarity (distance)

class GraphHellingerDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Hellinger)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Hellinger distance 
        helli = 1 - (np.sqrt(first * second)).sum()

        return  2.0 * sqrt(helli) #returns the dissimilarity (distance)

class GraphMatusitaDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Matusita)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Matusita distance 
        matu = (np.sqrt(first * second)).sum()

        return  sqrt(2.0 - 2.0 * matu) #returns the dissimilarity (distance)

class GraphSquaredchordDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Squared-chord)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Squared-chord distance (= Matusita without the square root)
        sqchord = (np.sqrt(first * second)).sum()

        return  2.0 - 2.0 * sqchord  #returns the dissimilarity (distance)

###################################################################
# Squared L_2 (or Chi^2) family (7 different distance metrics)
###################################################################

class GraphPearsonDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Pearson Chi^2)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        num =  (first - second) 
        num = num * num
        
        # Check possible division by zero
        almostzero = np.finfo(float).eps
        zero = 0.0
        second = np.where((second < almostzero), almostzero, second)       

        # Compute the Pearson Chi^2 distance
        pearson = np.where((num < almostzero), zero, num / second)
        
        
        return  pearson.sum() #returns the dissimilarity (distance)

class GraphNeymanDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Neyman Chi^2)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        num =  (first - second) 
        num = num * num

        # Check possible division by zero
        almostzero = np.finfo(float).eps
        zero = 0.0
        first = np.where((first < almostzero),  almostzero, first)
  
        # Compute the Neyman Chi^2 distance
        neyman = np.where((num < almostzero), zero, num / first)
               
        return  neyman.sum() #returns the dissimilarity (distance)

class GraphSquaredDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Squared Chi^2)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Squared Chi^2 distance
        num =  (first - second) 
        den = (first + second)
        
        # Check possible case of 0 / 0 
        squared = np.divide(num*num, den, out=np.zeros_like(den), where=(den!=0))
                
        return  squared.sum() #returns the dissimilarity (distance)

class GraphProbsymmetricDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Probabilistic symmetric Chi^2)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Probabistic symmetric Chi^2 distance
        num =  (first - second) 
        den = (first + second)

        # Check possible case of 0 / 0 
        probsym = np.divide(num*num, den, out=np.zeros_like(den), where=(den!=0))
        
        return  2.0 * probsym.sum() #returns the dissimilarity (distance)

class GraphDivergenceDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic:  PDF-based dissimilarity (Divergence)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Divergence
        num =  (first - second) 
        den = (first + second)
        
        # Check possible case of 0 / 0
        diverg = np.divide(num*num, den*den, out=np.zeros_like(den), where=(den!=0))
                
        return  2.0 * diverg.sum() #returns the dissimilarity (distance)

class GraphClarkDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Clark)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Clark distance
        num =  abs(first - second) 
        den = (first + second)
        
        # Check possible case of 0 / 0
        clark = np.divide(num,den, out=np.zeros_like(den), where=(den!=0))
        clark = (clark*clark).sum()
        
        return  sqrt(clark) #returns the dissimilarity (distance)

class GraphAdditivesymmetricDissimilarity(GraphComparator):

    def compare_graphs(self):
        """
        Heuristic: PDF-based dissimilarity (Additive symmetric Chi^2)
        The two arrays are assumed
        --  of the same length (minimum of the length is chosen)
        --  corresponding to the same ordered set of nodes:
            graph resizing has to be done before!
        """

        # Matrices normalization
        first, second = self._normalize_matrices()

        # Compute the Additive symmetric Chi^2 distance
        term1 =  (first - second) 
        term2 =  (first + second)
        num = term1 * term1 * term2
        den = (first * second)
        # Check possible case of 0 / 0
        addsym = np.divide(num,den, out=np.zeros_like(den), where=(den!=0))
        
        return  addsym.sum() #returns the dissimilarity (distance)
